import time
import threading
from bme68x import BME68X
# TODO: fixme -> these are now part of bme68x
import com_picockpit.bme688airmeter.bme68xConstants as BME68XCONST
import com_picockpit.bme688airmeter.bsecConstants as BSECCONST

#import com_picockpit.bme688airmeter.constants as CONST
from picockpit_client.pmqtt import PCPMQTTClient as PCPMQTTClient


class ComPicockpitBME688Airmeter:
    def __init__(self,
                 global_run_flag,
                 resource_manager,
                 hooks):
        self.global_run_flag = global_run_flag
        self.hooks = hooks
        self.topic_app_prefix = "com.picockpit/bme688airmeter/"
        self.st = threading.Thread(target=self.stats_thread)
        self.run_flag = threading.Event()
        self.resource_manager = resource_manager
        self.on = False
        self.addr = BME68XCONST.BME68X_I2C_ADDR_HIGH
        # advanced settings
        self.adv = False
        self.sensor = None
        # do we enable the iaq processing?
        self.iaq = False
        self.error = False
        self.error_message = ""
        # run_flag controls the execution of the thread;
        # it is set in the on_connect_hook!
        # self.run_flag.set()
        self.bme_is_initialized = False
        self.keep_alive = int(time.time())
    
    def save_bsec_state(self):
        state_file_path = "/var/spool/picockpit-client/apps/" + self.topic_app_prefix + "bsec_state.state"
        if not self.sensor:
            self.error_message = "Sensor needs to be initialized before state can be saved"
            self.publish_error()
            return False
        try:
            state = self.sensor.get_bsec_state()
            with open(state_file_path, 'w') as state_file:
                state_file.writelines('%i\n' % param for param in state)
            return True
        except:
            return False


    def load_bsec_state(self):
        state_file_path = "/var/spool/picockpit-client/apps/" + self.topic_app_prefix + "bsec_state.state"
        if not self.sensor:
            self.error_message = "Sensor needs to be initialized before state can be loaded"
            self.publish_error()
            return False
        try:
            with open(state_file_path, 'r') as state_file:
                state = list(map(int, state_file.readlines()))
        except:
            state_file = None
        if state_file:
            self.sensor.set_bsec_state(state)
            return True
        return False

    def sigterm_hook(self):
        # allow the internal thread which might be blocked to join.
        self.run_flag.set()

    def subscribe(self, topic, message_callback_function):
        self.hooks["subscribe"](
            topic=self.topic_app_prefix + topic,
            message_callback_function=message_callback_function)

    def unsubscribe(self, topic):
        self.hooks["unsubscribe"](
            topic=self.topic_app_prefix + topic)

    def publish(self, topic, message, retain=False, overwrite=True):
        self.hooks["publish"](
            topic_app_prefix=self.topic_app_prefix,
            topic_proper=topic,
            message=message,
            retain=retain,
            overwrite=overwrite)

    def publish_error(self):
        msg = {
            "$error": self.error_message
        }
        self.publish(
            topic='$state',
            message=msg,
            retain=False,
            overwrite=True)
        # force a delay of 4 seconds to slow down messages
        time.sleep(4)

    def check_keep_alive(self, timeout_sec):
        # block until run_flag is true
        # (we have just reconnected or have received a keep alive)
        # this run_flag will also be unblocked on the sigterm_hook being called
        # allow the tread to terminate by falling out at the bottom
            # self.run_flag.wait()
        # time.time() returns number of seconds since the epoch
        # using int to cast to an integer without fractions of a second
            # local.ts_last_keep_alive = int(time.time())
            # local.ts_now = int(time.time())
        # three minutes = 180 seconds
        return (int(time.time()) - self.keep_alive) < timeout_sec

    def stats_thread(self):
        ld = threading.local()
        while self.check_keep_alive(180) and self.global_run_flag.is_set() and self.on:
            try:
                # the flag will be set to true by the initial setup
                # and by periodic pings from the picockpit frontend
                # whenever it is set to true, we reset the last_keep_alive
                # and clear the flag (set it to false)
                # this way, if no further keep alives arrive, the system
                # will drop into blocking / waiting mode at some point
                if self.run_flag.is_set():
                    self.keep_alive = int(time.time())
                    self.run_flag.clear()
                if self.on:
                    if self.bme_is_initialized:
                        if self.resource_manager.i2c.i2c_active:
                            try:
                                # time.sleep(1)
                                # if I2C ADDR self.addr is blocked we need to wait until it is available
                                stats=self.get_stats()
                                while not stats:
                                    stats=self.get_stats()
                                self.publish(
                                    topic='$state',
                                    message=stats,
                                    retain=False,
                                    overwrite=True)
                            except TypeError:
                                self.error_message = "I2C is not active - please enable it"
                                self.publish_error()
                        else:
                            self.error_message = "I2C is not active - please enable it"
                            self.publish_error()
                            # NB: self.error refers to a general sensor error - not I2C error
                            # therefore we do not set it here.
                    elif self.error:
                        self.publish_error()
                time.sleep(3)
                ld.ts_now = int(time.time())
            # self.save_bsec_state()
            except KeyboardInterrupt:
                break

    def start_stats_thread(self):
        try:
            self.sensor = BME68X(self.addr)
            # Saving default BSEC state
            bsec_default_state = self.sensor.get_bsec_state()
            with open('/var/spool/picockpit-client/apps/com.picockpit/bme688digitalnose/bsec_default_state.state', 'w') as state_file:
                state_file.writelines('%i\n' % param for param in bsec_default_state)
            # Open last BSEC state if existing
            self.load_bsec_state()
            # BSEC_SAMPLE_RATE_HIGH_PERFORMANCE = 0.055556
            BSEC_SAMPLE_RATE_LP = 0.33333
            self.sensor.set_sample_rate(BSEC_SAMPLE_RATE_LP)
            self.error = False
            self.error_message = ""
            self.bme_is_initialized = True
        except:
            self.error = True
            self.bme_is_initialized = False
            self.error_message = "Could not initialize BME68X - please check the address, and whether it is attached correctly!"
            # publish error
            self.publish_error()
        self.resource_manager.i2c.acquire_i2c_addr(self.addr, self.topic_app_prefix)
        self.on = True
        self.st.start()

    def kill_stats_thread(self):
        self.save_bsec_state()
        self.on = False
        self.st.join()
        self.st = threading.Thread(target=self.stats_thread)
        self.resource_manager.i2c.release_i2c_addr(self.addr, self.topic_app_prefix)

    # returns a Dict with the values in it, which can be published directly
    # TODO: rewrite for BME688 Airmeter
    def get_stats(self):
        stats = {}
        data = {}
        try:
            data = self.sensor.get_bsec_data()
        except Exception as e:
            print(e)
            self.error_message = "error while reading from the BME68X sensor!"
            self.publish_error()
            return stats
        if data == None or data == {}:
            time.sleep(0.5)
            return stats
        else:
            stats['ns'] = data['sample_nr']
            stats['ts'] = data['timestamp']

            if self.iaq:
                stats["i"] = data["iaq"]
                stats["ia"] = data["iaq_accuracy"]
                stats['rt'] = data['raw_temperature']
                stats['rh'] = data['raw_humidity']
                stats['rg'] = data['raw_gas']/1000
                stats['p'] = data['raw_pressure']/100

                if self.adv:
                    stats['h'] = data['humidity']
                    stats["t"] = data['temperature']
                    stats["si"] = data["static_iaq"]
                    stats["sia"] = data["static_iaq_accuracy"]
                    stats["c"] = data["co2_equivalent"]
                    stats["ca"] = data["co2_accuracy"]
                    stats["b"] = data["breath_voc_equivalent"]
                    stats["ba"] = data["breath_voc_accuracy"]
                    stats["g"] = data["comp_gas_value"]
                    stats["ga"] = data["comp_gas_accuracy"]
                    stats["gp"] = data["gas_percentage"]
                    stats["gpa"] = data["gas_percentage_accuracy"]
            else:
                stats['s'] = data['status']
                stats['rt'] = data['temperature']
                stats['p'] = data['pressure']
                stats['rh'] = data['humidity']
                stats['rg'] = data['gas_resistance']

        return stats

    def timer_hook(self, timer_spacing):
        pass

    def on_set_msg(self, client, userdata, message):
        if self.st.is_alive():
            # kill stats thread to restart cleanly
            self.kill_stats_thread()
        topic, payload = PCPMQTTClient.parse_message(message)
        # set self.on to True if msg requests it, so we can set it to False if any condition is not fullfilled
        if "on" in payload and payload["on"] is True:
            self.on = True
        # check if we need to switch i2c address and if the final address is available
        if "addr" in payload:
            if payload["addr"]:
                self.addr = BME68XCONST.BME68X_I2C_ADDR_HIGH
            else:
                self.addr = BME68XCONST.BME68X_I2C_ADDR_LOW
            if self.resource_manager.i2c.is_locked(self.addr):
                self.on = False
                self.error = True
                self.error_message = f"I2C Address {hex(self.addr)} is already in use by {self.resource_manager.i2c.get_owner(self.addr)} - Please choose another I2C Address"
                self.publish_error()
        else:
            self.on = False
            self.error = True
            self.error_message = "No I2C Address provided - Please choose an I2C Address"
            self.publish_error()
        # check if advanced mode shall be enabled
        if "adv" in payload:
            old_adv = self.adv
            self.adv = payload["adv"]
            if old_adv != self.adv:
                self.publish_meta()
        need_bsec_switch = False
        # check if IAQ shall be enabled
        if "iaq" in payload:
            old_iaq = self.iaq
            self.iaq = payload["iaq"]
            if old_iaq != self.iaq:
                self.publish_meta()
                # switch BSEC
                need_bsec_switch = True
        # if still self.on then we start te stats thread
        if self.on:
            self.start_stats_thread()

    def on_control_msg(self, client, userdata, message):
        # this resets the thread's countdown.
        # not interested in the message contents
        self.run_flag.set()
        self.keep_alive = int(time.time())

    def on_set_bsec_state(self, client, userdata, message):
        topic, payload = PCPMQTTClient.parse_message(message)
        print(topic, payload)
        try:
            if "$reset" in payload:
                if payload["$reset"]:
                    print("NEED TO RESET BSEC STATE HERE")
                    # TODO
                    # with open('/var/spool/picockpit-client/apps/com.picockpit/bme688digitalnose/bsec_default_state.state', 'r') as state_file:
                    #     bsec_state = list(map(int, state_file.readlines()))
            else:
                print('BSEC STATE UPDATE REQUESTED')
        except Exception:
            self.error = True
            self.error_message = "Failed to switch BSEC state"
            self.publish_error()

    def reload_hook(self):
        # avoid double subscription (possibly unnecessary here)
        self.unsubscribe(topic="control")
        self.unsubscribe(topic="set")
        # this possibly should ensure that the client is connected ..
        # or better: ensure upstream that no errors are generated by this
        self.on_connect_hook()

    def publish_meta(self):
        # these parameters are always present
        meta_base = {
                "ns": {
                    "$description": "sample number",
                    "$unit": ""
                },
                "ts": {
                    "$description": "timestamp",
                    "$unit": "ms"
                },
                "rt": {
                    "$description": "air temperature (raw)",
                    "$unit": "°C",
                    "$min": -40,
                    "$max": 85
                },
                "p": {
                    "$description": "air pressure",
                    "$unit": "hPa",
                    "$min": 300,
                    "$max": 1100
                },
                "rh": {
                    "$description": "air humidity (raw)",
                    "$unit": "%",
                    "$min": 0,
                    "$max": 100
                },
                "rg": {
                    "$description": "gas resistance (raw)",
                    "$unit": "kΩ"
                }            
        }

        meta_status = {
                "s": {
                    "$description": "status code",
                    "$unit": ""
                },
        }

        meta_iaq = {
                "i": {
                    "$description": "air quality (IAQ)",
                    "$unit": ""
                }        
        }

        meta_adv = {
                "ia": {
                    "$description": "air quality (IAQ) accuracy",
                    "$unit": ""
                },
                "t": {
                    "$description": "air temperature",
                    "$unit": "°C",
                    "$min": -40,
                    "$max": 85
                },
                "h": {
                    "$description": "air humidity",
                    "$unit": "%",
                    "$min": 0,
                    "$max": 100
                },
                "si": {
                    "$description": "static IAQ (raw)",
                    "$unit": ""
                },
                "sia": {
                    "$description": "static IAQ accuracy",
                    "$unit": ""
                },
                "c": {
                    "$description": "CO2 equivalent",
                    "$unit": ""
                },
                "ca": {
                    "$description": "CO2 equivalent accuracy",
                    "$unit": ""
                },
                "b": {
                    "$description": "breath VOC equivalent",
                    "$unit": ""
                },
                "ba": {
                    "$description": "breath VOC equivalent accuracy",
                    "$unit": ""
                },
                "g": {
                    "$description": "comp gas value",
                    "$unit": ""
                },
                "ga": {
                    "$description": "comp gas value accuracy",
                    "$unit": ""
                },
                "gp": {
                    "$description": "gas percentage",
                    "$unit": ""
                },
                "gpa": {
                    "$description": "gas percentage accuracy",
                    "$unit": ""
                }
        }

        if self.iaq:
            if self.adv:
                meta = {**meta_base, **meta_iaq, **meta_adv}
            else:
                meta = {**meta_base, **meta_iaq}
        else: 
            meta = {**meta_base, **meta_status}
        self.publish(
            topic='$meta',
            message=meta,
            retain=True,
            overwrite=True)

    def on_connect_hook(self):
        self.subscribe(
            topic="control",
            message_callback_function=self.on_control_msg)
        self.subscribe(
            topic="set",
            message_callback_function=self.on_set_msg)
        self.publish_meta()
