#!/usr/bin/python3
import argparse
from picockpit_client.pcc import PiCockpitClient
from picockpit_client.daemon_mode import DaemonMode
from picockpit_client.interactive_mode import InteractiveMode

SETTINGS_FILE = '/etc/picockpit-client/picockpit-client.config.json'


def main():
    # read command line options
    parser = argparse.ArgumentParser(
        description='client for picockpit.com',
        prog='picockpit-client',
        epilog='Currently available commands: connect | convertcreds | daemon | getprefix | help | info | setupmqtt | version')
    parser.add_argument(
        'command',
        help="command to execute. Currently only 'connect', 'convertcreds', 'daemon', 'getprefix', 'help', 'info', 'setupmqtt' and 'version' are supported")
    parser.add_argument(
        '--verbose',
        help="increase verbosity where applicable.",
        action='store_true')
    parser.add_argument(
        '--serial',
        help="override serial of this pi (not recommended!)")
    parser.add_argument(
        '--apikey',
        help="pass api key on command line (only recommended for scripting!)")
    parser.add_argument(
        '--yes',
        help="assume yes to all questions, proceed automatically",
        action="store_true")
    parser.add_argument(
        '--skipkey',
        help="Skip adding the key for connect (requires a key to be present already)",
        action='store_true')
    parser.add_argument(
        '--name',
        help="Set the name for the pi to be added")
    # parser.add_argument(
    #     '--privacy',
    #     help="Run PiDoctor in privacy mode",
    #     action='store_true')
    parser.add_argument(
        '--selfsigned',
        help="Allow self signed certificates for MQTT server",
        action='store_true')

    args = parser.parse_args()
    if args.command == 'daemon':
        pcp = PiCockpitClient(
            config_file_path=SETTINGS_FILE,
            args=args,
            daemon_mode=True)
        DaemonMode(args=args, pcp=pcp, verbose=pcp.verbose)
    elif args.command == 'help':
        parser.print_help()
        print("\nHave a nice day!")
    else:
        pcp = PiCockpitClient(
            config_file_path=SETTINGS_FILE,
            args=args,
            daemon_mode=False)        
        InteractiveMode(args=args, pcp=pcp, verbose=pcp.verbose)


if __name__ == '__main__':
    main()
